//#include <linux/config.h>
//#include <linux/types.h>
//#include <linux/string.h>
//#include <linux/sched.h>

#include <asm/i387.h>
//#include <asm/hardirq.h> 


/*
 *	MMX/3DNow! library helper functions
 *
 *	To do:
 *	We can use MMX just for prefetch in IRQ's. This may be a win. 
 *		(reported so on K6-III)
 *	We should use a better code neutral filler for the short jump
 *		leal ebx. [ebx] is apparently best for K6-2, but Cyrix ??
 *	We also want to clobber the filler register so we don't get any
 *		register forwarding stalls on the filler. 
 *
 *	Add *user handling. Checksums are not a win with MMX on any CPU
 *	tested so far for any MMX solution figured.
 *
 *	22/09/2000 - Arjan van de Ven 
 *		Improved for non-egineering-sample Athlons 
 *
 */
 
/*
 *	The K7 has streaming cache bypass load/store. The Cyrix III, K6 and
 *	other MMX using processors do not.
 */

void zero_page_3dnow(void *page)
{
	int i;

	kernel_fpu_begin();
	
	__asm__ __volatile__ (
		"	pxor %%mm0, %%mm0\n" : :
	);

	for(i = 0; i < PAGE_SIZE/64; i++)
	{
		__asm__ __volatile__ (
		"	movntq %%mm0, (%0)\n"
		"	movntq %%mm0, 8(%0)\n"
		"	movntq %%mm0, 16(%0)\n"
		"	movntq %%mm0, 24(%0)\n"
		"	movntq %%mm0, 32(%0)\n"
		"	movntq %%mm0, 40(%0)\n"
		"	movntq %%mm0, 48(%0)\n"
		"	movntq %%mm0, 56(%0)\n"
		: : "r" (page) : "memory"
		);
		page+=64;
	}
	/* since movntq is weakly-ordered, a "sfence" is needed to become
	 * ordered again.
	 */
	__asm__ __volatile__ (
		"	sfence\n" : :
	);
	kernel_fpu_end();
}

void copy_page_3dnow(void *to, void *from)
{
	int i;

	kernel_fpu_begin();

	/* maybe the prefetch stuff can go before the expensive fnsave...
	 * but that is for later. -AV
	 */
	__asm__ __volatile__ (
		"1:	prefetch (%0)\n"
		"	prefetch 64(%0)\n"
		"	prefetch 128(%0)\n"
		"	prefetch 192(%0)\n"
		"	prefetch 256(%0)\n"
		"2:\n"
		".section .fixup, \"ax\"\n"
		"3:	movw $0x1AEB, 1b\n"	/* jmp on 26 bytes */
		"	jmp 2b\n"
		".previous\n"
		".section __ex_table,\"a\"\n"
		"	.align 4\n"
		"	.long 1b, 3b\n"
		".previous"
		: : "r" (from)
	);

	for(i = 0; i < (PAGE_SIZE-320)/64; i++)
	{
		__asm__ __volatile__ (
		"1:	prefetch 320(%0)\n"
		"2:	movq (%0), %%mm0\n"
		"	movntq %%mm0, (%1)\n"
		"	movq 8(%0), %%mm1\n"
		"	movntq %%mm1, 8(%1)\n"
		"	movq 16(%0), %%mm2\n"
		"	movntq %%mm2, 16(%1)\n"
		"	movq 24(%0), %%mm3\n"
		"	movntq %%mm3, 24(%1)\n"
		"	movq 32(%0), %%mm4\n"
		"	movntq %%mm4, 32(%1)\n"
		"	movq 40(%0), %%mm5\n"
		"	movntq %%mm5, 40(%1)\n"
		"	movq 48(%0), %%mm6\n"
		"	movntq %%mm6, 48(%1)\n"
		"	movq 56(%0), %%mm7\n"
		"	movntq %%mm7, 56(%1)\n"
		".section .fixup, \"ax\"\n"
		"3:	movw $0x05EB, 1b\n"	/* jmp on 5 bytes */
		"	jmp 2b\n"
		".previous\n"
		".section __ex_table,\"a\"\n"
		"	.align 4\n"
		"	.long 1b, 3b\n"
		".previous"
		: : "r" (from), "r" (to) : "memory"
		);
		from+=64;
		to+=64;
	}
	for(i = (PAGE_SIZE-320)/64; i < PAGE_SIZE/64; i++)
	{
		__asm__ __volatile__ (
		"2:	movq (%0), %%mm0\n"
		"	movntq %%mm0, (%1)\n"
		"	movq 8(%0), %%mm1\n"
		"	movntq %%mm1, 8(%1)\n"
		"	movq 16(%0), %%mm2\n"
		"	movntq %%mm2, 16(%1)\n"
		"	movq 24(%0), %%mm3\n"
		"	movntq %%mm3, 24(%1)\n"
		"	movq 32(%0), %%mm4\n"
		"	movntq %%mm4, 32(%1)\n"
		"	movq 40(%0), %%mm5\n"
		"	movntq %%mm5, 40(%1)\n"
		"	movq 48(%0), %%mm6\n"
		"	movntq %%mm6, 48(%1)\n"
		"	movq 56(%0), %%mm7\n"
		"	movntq %%mm7, 56(%1)\n"
		: : "r" (from), "r" (to) : "memory"
		);
		from+=64;
		to+=64;
	}
	/* since movntq is weakly-ordered, a "sfence" is needed to become
	 * ordered again.
	 */
	__asm__ __volatile__ (
		"  sfence\n" : :
	);
	kernel_fpu_end();
}

/*
 *	Generic MMX implementation without K7 specific streaming
 */
void zero_page_mmx(void *page)
{
	int i;
	
	kernel_fpu_begin();
	
	__asm__ __volatile__ (
		"	pxor %%mm0, %%mm0\n" : :
	);

	for(i = 0; i < PAGE_SIZE/128; i++)
	{
		__asm__ __volatile__ (
		"	movq %%mm0, (%0)\n"
		"	movq %%mm0, 8(%0)\n"
		"	movq %%mm0, 16(%0)\n"
		"	movq %%mm0, 24(%0)\n"
		"	movq %%mm0, 32(%0)\n"
		"	movq %%mm0, 40(%0)\n"
		"	movq %%mm0, 48(%0)\n"
		"	movq %%mm0, 56(%0)\n"
		"	movq %%mm0, 64(%0)\n"
		"	movq %%mm0, 72(%0)\n"
		"	movq %%mm0, 80(%0)\n"
		"	movq %%mm0, 88(%0)\n"
		"	movq %%mm0, 96(%0)\n"
		"	movq %%mm0, 104(%0)\n"
		"	movq %%mm0, 112(%0)\n"
		"	movq %%mm0, 120(%0)\n"
		: : "r" (page) : "memory"
		);
		page+=128;
	}

	kernel_fpu_end();
}

void copy_page_mmx(void *to, void *from)
{
	int i;
	
	
	kernel_fpu_begin();

	__asm__ __volatile__ (
		"1:	prefetch (%0)\n"
		"	prefetch 64(%0)\n"
		"	prefetch 128(%0)\n"
		"	prefetch 192(%0)\n"
		"	prefetch 256(%0)\n"
		"2:\n"
		".section .fixup, \"ax\"\n"
		"3: movw $0x1AEB, 1b\n"	/* jmp on 26 bytes */
		"	jmp 2b\n"
		".previous\n"
		".section __ex_table,\"a\"\n"
		"	.align 4\n"
		"	.long 1b, 3b\n"
		".previous"
		: : "r" (from)
	);

	for(i = 0; i < PAGE_SIZE/64; i++)
	{
		__asm__ __volatile__ (
		"1:	prefetch 320(%0)\n"
		"2:	movq (%0), %%mm0\n"
		"	movq 8(%0), %%mm1\n"
		"	movq 16(%0), %%mm2\n"
		"	movq 24(%0), %%mm3\n"
		"	movq %%mm0, (%1)\n"
		"	movq %%mm1, 8(%1)\n"
		"	movq %%mm2, 16(%1)\n"
		"	movq %%mm3, 24(%1)\n"
		"	movq 32(%0), %%mm0\n"
		"	movq 40(%0), %%mm1\n"
		"	movq 48(%0), %%mm2\n"
		"	movq 56(%0), %%mm3\n"
		"	movq %%mm0, 32(%1)\n"
		"	movq %%mm1, 40(%1)\n"
		"	movq %%mm2, 48(%1)\n"
		"	movq %%mm3, 56(%1)\n"
		".section .fixup, \"ax\"\n"
		"3:	movw $0x05EB, 1b\n"	/* jmp on 5 bytes */
		"	jmp 2b\n"
		".previous\n"
		".section __ex_table,\"a\"\n"
		"	.align 4\n"
		"	.long 1b, 3b\n"
		".previous"
		: : "r" (from), "r" (to) : "memory"
		);
		from+=64;
		to+=64;
	}
	kernel_fpu_end();
}
