#include <linux/kernel.h>
#include <linux/sched.h>

#include <asm/hardirq.h> 

#include <linux/sysctl.h>

//#include <linux/config.h>
//#include <linux/types.h>
//#include <linux/string.h>
//#include <asm/i387.h>

#define DECL_OP(a) \
	void zero_page_##a(void*); \
	void copy_page_##a(void*, const void*)


/*
 */
static void zero_page_slow(void * page)
{
	int d0, d1;
	__asm__ __volatile__(
		"	cld\n"
		"	rep ; stosl\n"
		: "=&c" (d0), "=&D" (d1)
		:"a" (0),"1" (page),"0" (PAGE_SIZE/4)
		:"memory"
	);
}
 
static void copy_page_slow(void *to, const void *from)
{
	int d0, d1, d2;
	__asm__ __volatile__(
		"	cld\n"
		"	rep ; movsl\n"
		: "=&c" (d0), "=&D" (d1), "=&S" (d2)
		: "0" (PAGE_SIZE/4),"1" ((long) to),"2" ((long) from)
		: "memory"
	);
}

int change_pageops = 0;

static void (*zero_f)(void *) = zero_page_slow;
static void (*copy_f)(void *, const void*) = copy_page_slow;

static int zero_cnt,copy_cnt;

DECL_OP(mmx);
DECL_OP(3dnow);
DECL_OP(sse);
DECL_OP(sse_apn); DECL_OP(mmx_apn); DECL_OP(gp_apn);
DECL_OP(sse_Apn); DECL_OP(mmx_Apn); DECL_OP(gp_Apn);
DECL_OP(sse_aPn); DECL_OP(mmx_aPn); DECL_OP(gp_aPn);
DECL_OP(sse_APn); DECL_OP(mmx_APn); DECL_OP(gp_APn);
DECL_OP(sse_apN); DECL_OP(mmx_apN); DECL_OP(gp_apN);
DECL_OP(sse_ApN); DECL_OP(mmx_ApN); DECL_OP(gp_ApN);
DECL_OP(sse_aPN); DECL_OP(mmx_aPN); DECL_OP(gp_aPN);
DECL_OP(sse_APN); DECL_OP(mmx_APN); DECL_OP(gp_APN);

#define SW_TO(a) do { \
	zero_f = zero_page_##a; \
	copy_f = copy_page_##a; \
	msg = #a; \
} while(0)

static void change_ops(void)
{
	const char *msg = 0;

	switch(change_pageops) {
	case 1: SW_TO(slow); break;
	case 2: SW_TO(mmx); break;
	case 3: SW_TO(3dnow); break;
	case 4: SW_TO(sse); break;
	case 5:
		zero_f = zero_page_mmx_APn;
		copy_f = copy_page_mmx_APN;
		msg = "mmx_APn/APN";
		break;
	case 6:
		zero_f = zero_page_sse_APn;
		copy_f = copy_page_sse_APN;
		msg = "sse_APn/APN";
		break;
	case 10003: SW_TO(sse_apn); break; case 10002: SW_TO(mmx_apn); break; case 10001: SW_TO(gp_apn); break; 
	case 11003: SW_TO(sse_Apn); break; case 11002: SW_TO(mmx_Apn); break; case 11001: SW_TO(gp_Apn); break; 
	case 10103: SW_TO(sse_aPn); break; case 10102: SW_TO(mmx_aPn); break; case 10101: SW_TO(gp_aPn); break; 
	case 11103: SW_TO(sse_APn); break; case 11102: SW_TO(mmx_APn); break; case 11101: SW_TO(gp_APn); break; 
	case 10013: SW_TO(sse_apN); break; case 10012: SW_TO(mmx_apN); break; case 10011: SW_TO(gp_apN); break; 
	case 11013: SW_TO(sse_ApN); break; case 11012: SW_TO(mmx_ApN); break; case 11011: SW_TO(gp_ApN); break; 
	case 10113: SW_TO(sse_aPN); break; case 10112: SW_TO(mmx_aPN); break; case 10111: SW_TO(gp_aPN); break; 
	case 11113: SW_TO(sse_APN); break; case 11112: SW_TO(mmx_APN); break; case 11111: SW_TO(gp_APN); break; 
	default:
		printk("unimplemented!\n");
	}
	if(msg) {
		printk("Switched to %s clear/copy page ops, %d/%d ops were done before\n",
			msg, zero_cnt, copy_cnt);
		zero_cnt = copy_cnt = 0;
	}
	change_pageops = 0;
}

void clear_page(void *page)
{
	zero_cnt++;
	if(unlikely(in_interrupt())) {
		zero_page_slow(page);
		return;
	}
	if(!change_pageops) {
		zero_f(page);
		return;
	}
	change_ops();
	zero_f(page);
}

void copy_page(void *to, const void *from)
{
	copy_cnt++;
	if(unlikely(in_interrupt())) {
		copy_page_slow(to, from);
		return;
	}
	if(!change_pageops) {
		copy_f(to, from);
		return;
	}
	change_ops();
	copy_f(to, from);
}

static struct ctl_table pageop_table[] = {
	{
		.ctl_name	= 19847,	/* I typed random number */
		.procname	= "pageop",
		.data		= &change_pageops,
		.maxlen		= sizeof(change_pageops),
		.mode		= 0644,
		.proc_handler	= &proc_dointvec,
	},
	{ .ctl_name = 0 }
};

static int __init pageops_init(void)
{
	register_sysctl_table(pageop_table, 1);
	return 0;
}

module_init(pageops_init)
