/*
* linux/arch/i386/lib/sse.c
*
* Copyright 2004 Jens Maurer
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* Send feedback to <Jens.Maurer@gmx.net>
*/

//#include <linux/config.h>
//#include <linux/module.h>
//#include <linux/types.h>
//#include <linux/kernel.h>
//#include <linux/kernel.h>
#include <linux/preempt.h>	/* preempt_disable */
#include <asm/page.h>		/* PAGE_SIZE */
#include <asm/system.h>		/* cr0 ops */


/*
*	SSE library helper functions
*/

#define SSE_START(cr0) do { \
	preempt_disable(); \
	cr0 = read_cr0(); \
	clts(); \
	} while(0)


#define SSE_END(cr0) do { \
	write_cr0(cr0); \
	preempt_enable(); \
	} while(0)

void zero_page_sse(void * page)
{
	unsigned char xmm_save[16];
	unsigned int cr0;
	int i;

	SSE_START(cr0);
	asm volatile(
		"	movups %%xmm0, (%0)\n"
		"	xorps %%xmm0, %%xmm0\n"
		: : "r" (xmm_save)
	);
	for(i = 0; i < PAGE_SIZE/16/4; i++) {
		asm volatile(
			"	movntps %%xmm0,   (%0)\n"
			"	movntps %%xmm0, 16(%0)\n"
			"	movntps %%xmm0, 32(%0)\n"
			"	movntps %%xmm0, 48(%0)\n"
			: : "r"(page) : "memory"
		);
		page += 16*4;
	}
	asm volatile(
		"	movups (%0), %%xmm0\n"
		"	sfence\n"
		: : "r" (xmm_save) : "memory"
	);
	SSE_END(cr0);
}

void copy_page_sse(void *to, void *from)
{
	unsigned char xmm_save[16*4];
	unsigned int cr0;
	int i;

	SSE_START(cr0);
	asm volatile(
		"	movups %%xmm0,   (%0)\n"
		"	movups %%xmm1, 16(%0)\n"
		"	movups %%xmm2, 32(%0)\n"
		"	movups %%xmm3, 48(%0)\n"
		: : "r" (xmm_save)
	);
	for(i = 0; i < PAGE_SIZE/16/4; i++) {
		asm volatile(
			"	movaps   (%0), %%xmm0\n"
			"	movaps 16(%0), %%xmm1\n"
			"	movaps 32(%0), %%xmm2\n"
			"	movaps 48(%0), %%xmm3\n"
			"	movntps %%xmm0,   (%1)\n"
			"	movntps %%xmm1, 16(%1)\n"
			"	movntps %%xmm2, 32(%1)\n"
			"	movntps %%xmm3, 48(%1)\n"
			: : "r" (from), "r" (to) : "memory"
		);
		from += 16*4;
		to += 16*4;
	}
	asm volatile(
		"	movups   (%0), %%xmm0\n"
		"	movups 16(%0), %%xmm1\n"
		"	movups 32(%0), %%xmm2\n"
		"	movups 48(%0), %%xmm3\n"
		"	sfence\n"
		: : "r" (xmm_save) : "memory"
	);
	SSE_END(cr0);
}
